//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
#property description "Portfolio-Combinator"
#property description "Transcendreamer"

#include <Graphics\Graphic.mqh>
#include <Math\Alglib\alglib.mqh>

input int random_sets=100000;
enum axis {MEAN,STDEV,SHARPE,PROFIT,LOSS,PF,R2};
input axis x_axis=STDEV;
input axis y_axis=MEAN;
input string symbol_name="PORTFOLIO";

CGraphic graph;

int datapoints=0;
int datarows=0;

struct DATA { double equity[]; datetime time[]; };
DATA dataset[];

double trend_step=100;

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int OnInit(void)
  {
   if(Loading())
     {
      Generation();
      Regression();
     }
   return(INIT_SUCCEEDED);
  }

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void OnDeinit(const int reason)
  {
   ChartSetInteger(0,CHART_SHOW,1);
   graph.Destroy();
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool Loading()
  {

   Print("###### DATA LOADING ######");

//---

   string filenames[];

   datarows=FileSelectDialog("SELECT EQUITY DATA FILES",NULL,
                             "CSV files (*.csv)|*.csv|All files (*.*)|*.*",
                             FSD_ALLOW_MULTISELECT|FSD_COMMON_FOLDER,filenames);

   if(datarows<1)
     {
      Alert("NO DATA FILES SELECTED");
      return false;
     }

   ArrayResize(dataset,datarows);

   for(int i=0; i<datarows; i++)
     {

      int handle=FileOpen(filenames[i],FILE_READ|FILE_IS_CSV|FILE_ANSI|FILE_COMMON,'\\');

      if(handle==INVALID_HANDLE)
        {
         Alert("CAN NOT OPEN FILE: ",filenames[i]);
         continue;
        }

      datapoints=0;

      while(!FileIsEnding(handle))
        {
         string parts[];
         string text=FileReadString(handle);
         int k=StringSplit(text,'\t',parts);

         datapoints++;
         ArrayResize(dataset[i].equity,datapoints,256);
         ArrayResize(dataset[i].time,datapoints,256);

         dataset[i].equity[datapoints-1]=StringToDouble(parts[1]);
         dataset[i].time[datapoints-1]=StringToTime(parts[0]);
        }

      FileClose(handle);

     }

//---

   int datasize[];
   ArrayResize(datasize,datarows);

   for(int i=0; i<datarows; i++)
      if(datasize[i]!=datasize[0])
        {
         Alert("INCORRECT DATA SIZES");
         return false;
        }

   for(int j=0; j<datapoints; j++)
      for(int i=0; i<datarows; i++)
         if(dataset[i].time[j]!=dataset[0].time[j])
           {
            Alert("INCORRECT TIME STAMPS");
            return false;
           }

   Print("LOADED FILES: ",datarows);
   Print("LOADED POINTS: ",datapoints);

   return true;

  }

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Generation()
  {

   Print("###### PORTFOLIO GENERATION ######");

//---

   double mean[];
   double stdev[];
   double sharpe[];
   double profit[];
   double loss[];
   double pf[];
   double r2[];

   ArrayResize(mean,random_sets);
   ArrayResize(stdev,random_sets);
   ArrayResize(sharpe,random_sets);
   ArrayResize(profit,random_sets);
   ArrayResize(loss,random_sets);
   ArrayResize(pf,random_sets);
   ArrayResize(r2,random_sets);

//---

   for(int k=0; k<random_sets; k++)
     {

      double weights[];
      ArrayResize(weights,datarows);
      ArrayInitialize(weights,0);

      for(int i=0; i<datarows; i++)
         weights[i]=(double)MathRand()/SHORT_MAX;

      double sum=0;
      for(int i=0; i<datarows; i++)
         sum+=weights[i];

      for(int i=0; i<datarows; i++)
         weights[i]=weights[i]/sum;

      //---

      double PORTFOLIO[];
      ArrayResize(PORTFOLIO,datapoints);
      ArrayInitialize(PORTFOLIO,0);

      for(int j=0; j<datapoints; j++)
         for(int i=0; i<datarows; i++)
            PORTFOLIO[j]+=
               dataset[i].equity[j]*weights[i];

      //---

      double x=0,y=0,x2=0,xy=0;

      for(int i=0; i<datapoints; i++)
        {
         x=x+i;
         y=y+PORTFOLIO[i];
         xy=xy+i*PORTFOLIO[i];
         x2=x2+i*i;
        }

      double slope=(datapoints*xy-x*y)/(datapoints*x2-x*x);
      double intercept=(y-slope*x)/datapoints;

      double MODEL[];
      ArrayResize(MODEL,datapoints);

      for(int i=0; i<datapoints; i++)
         MODEL[i]=slope*i+intercept;

      //---

      double DEV[];
      ArrayResize(DEV,datapoints);

      for(int j=0; j<datapoints; j++)
         DEV[j]=PORTFOLIO[j]-MODEL[j];

      double ssreg=0;
      double ssres=0;

      for(int j=0; j<datapoints; j++)
         ssreg+=(PORTFOLIO[j]*PORTFOLIO[j]);

      for(int j=0; j<datapoints; j++)
         ssres+=(DEV[j]*DEV[j]);

      r2[k]=ssreg/(ssreg+ssres);

      //---

      double change[];
      ArrayResize(change,datapoints);

      for(int j=0; j<datapoints; j++)
         if(j>0)
            change[j]=PORTFOLIO[j]-PORTFOLIO[j-1];
         else
            change[j]=0;

      //---

      mean[k]=0;
      for(int j=0; j<datapoints; j++)
         mean[k]+=change[j];
      mean[k]/=datapoints;

      stdev[k]=0;
      for(int j=0; j<datapoints; j++)
         stdev[k]+=(change[j]-mean[k])*(change[j]-mean[k]);
      stdev[k]=MathSqrt(stdev[k]/datapoints);

      sharpe[k]=0;
      if(stdev[k]!=0)
         sharpe[k]=mean[k]/stdev[k];

      //---

      profit[k]=0;
      loss[k]=0;

      for(int j=0; j<datapoints; j++)
         if(change[j]>0)
            profit[k]+=change[j];
         else
            loss[k]+=-change[j];

      pf[k]=0;
      if(loss[k]!=0)
         pf[k]=profit[k]/loss[k];

     }

//---

   ChartSetInteger(0,CHART_SHOW,0);

   int w=(int)ChartGetInteger(0,CHART_WIDTH_IN_PIXELS,0);
   int h=(int)ChartGetInteger(0,CHART_HEIGHT_IN_PIXELS,0);

   graph.Destroy();
   graph.Create(0,"GRAPH",0,10,10,w-10,h-10);

   double xxx[];
   double yyy[];

   ArrayResize(xxx,random_sets);
   ArrayResize(yyy,random_sets);

   for(int k=0; k<random_sets; k++)
     {

      if(x_axis==MEAN)
         xxx[k]=mean[k];
      if(x_axis==STDEV)
         xxx[k]=stdev[k];
      if(x_axis==SHARPE)
         xxx[k]=sharpe[k];
      if(x_axis==PROFIT)
         xxx[k]=profit[k];
      if(x_axis==LOSS)
         xxx[k]=loss[k];
      if(x_axis==PF)
         xxx[k]=pf[k];
      if(x_axis==R2)
         xxx[k]=r2[k];

      if(y_axis==MEAN)
         yyy[k]=mean[k];
      if(y_axis==STDEV)
         yyy[k]=stdev[k];
      if(y_axis==SHARPE)
         yyy[k]=sharpe[k];
      if(y_axis==PROFIT)
         yyy[k]=profit[k];
      if(y_axis==LOSS)
         yyy[k]=loss[k];
      if(y_axis==PF)
         yyy[k]=pf[k];
      if(y_axis==R2)
         yyy[k]=r2[k];

     }

   CCurve *curve1=graph.CurveAdd(xxx,yyy,clrBlue,CURVE_POINTS,"PORTFOLIO");
   curve1.PointsType(POINT_DIAMOND);
   curve1.PointsSize(2);
   curve1.Color(clrMagenta);
   graph.CurvePlotAll();
   graph.Update();

//---

   Print("GENERATED PORTFOLIOS: ",random_sets);

   return;

  }

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Regression()
  {

   Print("###### PORTFOLIO REGRESSION ######");

//---

   int info;
   CLinearModelShell LM;
   CLRReportShell AR;
   CMatrixDouble MATRIX(datapoints,datarows+1);

   double MODEL[];
   ArrayResize(MODEL,datapoints);

   if(trend_step==0)
     {
      Alert("ZERO TREND STEP");
      return;
     }

   for(int j=0; j<datapoints; j++)
      MODEL[j]=trend_step*j;

   for(int i=0; i<datarows; i++)
      for(int j=0; j<datapoints; j++)
         MATRIX[j].Set(i,
                       dataset[i].equity[j]-
                       dataset[i].equity[0]);

   for(int j=0; j<datapoints; j++)
      MATRIX[j].Set(datarows,MODEL[j]);

   CAlglib::LRBuildZ(MATRIX,datapoints,datarows,info,LM,AR);

   if(info<0)
     {
      Alert("REGRESSION MODEL ERROR");
      return;
     }

   double ROOTS[];
   ArrayResize(ROOTS,datarows);
   CAlglib::LRUnpack(LM,ROOTS,datarows);

//---

   double sum=0;
   for(int i=0; i<datarows; i++)
      sum+=(ROOTS[i]);

   for(int i=0; i<datarows; i++)
      ROOTS[i]=ROOTS[i]/sum;

   Print("ROOTS:");
   for(int i=0; i<datarows; i++)
      Print(ROOTS[i]);

   Print("-------------------------");

//---

   double PORTFOLIO[];
   ArrayResize(PORTFOLIO,datapoints);
   ArrayInitialize(PORTFOLIO,0);

   for(int j=0; j<datapoints; j++)
      for(int i=0; i<datarows; i++)
         PORTFOLIO[j]+=
            dataset[i].equity[0]+
            (dataset[i].equity[j]-dataset[i].equity[0])*ROOTS[i];

//---

   double x=0,y=0,x2=0,xy=0;

   for(int i=0; i<datapoints; i++)
     {
      x=x+i;
      y=y+PORTFOLIO[i];
      xy=xy+i*PORTFOLIO[i];
      x2=x2+i*i;
     }

   double slope=(datapoints*xy-x*y)/(datapoints*x2-x*x);
   double intercept=(y-slope*x)/datapoints;

   for(int i=0; i<datapoints; i++)
     {
      MODEL[i]=slope*i+intercept;
     }

//---

   double DEV[];
   ArrayResize(DEV,datapoints);

   for(int j=0; j<datapoints; j++)
      DEV[j]=PORTFOLIO[j]-MODEL[j];

   double ssreg=0;
   double ssres=0;

   for(int j=0; j<datapoints; j++)
      ssreg+=(PORTFOLIO[j]*PORTFOLIO[j]);

   for(int j=0; j<datapoints; j++)
      ssres+=(DEV[j]*DEV[j]);

   double r2=ssreg/(ssreg+ssres);
   Print("R^2=",r2);

//---

   double change[];
   ArrayResize(change,datapoints);

   for(int j=0; j<datapoints; j++)
      if(j>0)
         change[j]=PORTFOLIO[j]-PORTFOLIO[j-1];
      else
         change[j]=0;

   double mean=0;
   for(int j=0; j<datapoints; j++)
      mean+=change[j];
   mean/=datapoints;

   double stdev=0;
   for(int j=0; j<datapoints; j++)
      stdev+=(change[j]-mean)*(change[j]-mean);
   stdev=MathSqrt(stdev/datapoints);

   double sharpe=0;
   if(stdev!=0)
      sharpe=mean/stdev;

   Print("MEAN=",mean);
   Print("STDEV=",stdev);
   Print("SHARPE=",sharpe);

//---

   double profit=0;
   double loss=0;

   for(int j=0; j<datapoints; j++)
      if(change[j]>0)
         profit+=change[j];
      else
         loss+=-change[j];

   double pf=0;
   if(loss!=0)
      pf=profit/loss;

   Print("PROFITS=",profit);
   Print("LOSSES=",loss);
   Print("PF=",pf);

//---

   MqlRates rates[];
   ArrayResize(rates,datapoints);

   long begin=LONG_MAX;
   long end=LONG_MIN;

   for(int j=0; j<datapoints; j++)
     {
      rates[j].close=PORTFOLIO[j];
      rates[j].open=PORTFOLIO[j];
      rates[j].high=PORTFOLIO[j];
      rates[j].low=PORTFOLIO[j];
      rates[j].time=dataset[0].time[j];
      rates[j].real_volume=0;
      rates[j].tick_volume=0;
      rates[j].spread=0;

      if(dataset[0].time[j]>end)
         end=dataset[0].time[j];

      if(dataset[0].time[j]<begin)
         begin=dataset[0].time[j];
     }

//---

   CustomSymbolCreate(symbol_name);
   CustomSymbolSetInteger(symbol_name,SYMBOL_DIGITS,2);

   if(!SymbolSelect(symbol_name,true))
     {
      Alert("CAN NOT FIND SYMBOL: ",symbol_name);
      return;
     }

   datetime time1=D'1984.01.01 00:00';
   datetime time2=D'2077.01.01 00:00';
   CustomRatesDelete(symbol_name,time1,time2);

   int k=CustomRatesReplace(symbol_name,begin,end,rates);
   Print("TOTAL PORTFOLIO POINTS: ",k);

   long id=ChartOpen(symbol_name,PERIOD_D1);
   ChartSetInteger(id,CHART_MODE,CHART_LINE);

//---

   double xxx[];
   double yyy[];

   ArrayResize(xxx,1);
   ArrayResize(yyy,1);

   if(x_axis==MEAN)
      xxx[0]=mean;
   if(x_axis==STDEV)
      xxx[0]=stdev;
   if(x_axis==SHARPE)
      xxx[0]=sharpe;
   if(x_axis==PROFIT)
      xxx[0]=profit;
   if(x_axis==LOSS)
      xxx[0]=loss;
   if(x_axis==PF)
      xxx[0]=pf;
   if(x_axis==R2)
      xxx[0]=r2;

   if(y_axis==MEAN)
      yyy[0]=mean;
   if(y_axis==STDEV)
      yyy[0]=stdev;
   if(y_axis==SHARPE)
      yyy[0]=sharpe;
   if(y_axis==PROFIT)
      yyy[0]=profit;
   if(y_axis==LOSS)
      yyy[0]=loss;
   if(y_axis==PF)
      yyy[0]=pf;
   if(y_axis==R2)
      yyy[0]=r2;

   CCurve *curve2=graph.CurveAdd(xxx,yyy,clrRed,CURVE_POINTS,"REGRESSION");
   curve2.PointsType(POINT_SQUARE);
   curve2.PointsFill(true);
   curve2.PointsSize(5);
   curve2.Color(clrRed);
   graph.CurvePlotAll();
   graph.Update();

//---

   return;

  }

//+------------------------------------------------------------------+
